from .reflect import Method, MethodStub, Op


def get_method(method):
    """Annotation for a callable method specifying that it is a GET method."""
    return MethodStub(Method.Get, method)


def put_method(method):
    """Annotation for a callable method specifying that it is a PUT method."""
    return MethodStub(Method.Put, method)


def post_method(method):
    """Annotation for a callable method specifying that it is a POST method."""
    return MethodStub(Method.Post, method)


def delete_method(method):
    """Annotation for a callable method specifying that it is a DELETE method."""
    return MethodStub(Method.Delete, method)


def get_op(op):
    """Annotation for a callable function specifying that it is a GET :class:`Op`."""
    return Op.Get(op)


def put_op(op):
    """Annotation for a callable function specifying that it is a PUT :class:`Op`."""
    return Op.Put(op)


def post_op(op):
    """Annotation for a callable function specifying that it is a POST :class:`Op`."""
    return Op.Post(op)


def delete_op(op):
    """Annotation for a callable function specifying that it is a DELETE :class:`Op`."""
    return Op.Delete(op)




"""Tinychain `State`s, such as `Map`, `Tuple`, and `Op`."""

from . import reflect
from .ref import OpRef
from .util import *


class State(object):
    """
    A Tinychain state, such as a `Chain` or `Op` or `Value`.

    Do not subclass `State` directly. Use a more specific type instead.
    """

    __uri__ = URI("/state")

    def __init__(self, form):
        self.__form__ = form

        if isinstance(form, URI):
            self.__uri__ = form

        reflect.gen_headers(self)

    def __json__(self):
        return {str(uri(self)): [to_json(form_of(self))]}

    def __ref__(self, name):
        return self.__class__(URI(name))

    def __repr__(self):
        return f"{self.__class__.__name__}({form_of(self)})"

    def dtype(self):
        """Return the native `Class` of this `State`."""
        return Class(OpRef.get(uri(self) + "/class"))


class Map(State):
    """A key-value map whose keys are `Id`s and whose values are `State`s."""

    __uri__ = uri(State) + "/map"

    def __getitem__(self, key):
        return OpRef.Get(uri(self), key)

    def __json__(self):
        return to_json(form_of(self))


class Tuple(State):
    """A tuple of `State`s."""

    __uri__ = uri(State) + "/tuple"

    def __json__(self):
        return to_json(form_of(self))

    def __getitem__(self, key):
        return OpRef.Get(uri(self), key)


# Scalar types

class Scalar(State):
    """
    An immutable `State` which always resides entirely in the host's memory.

    Do not subclass `Scalar` directly. Use :class:`Value` instead.
    """

    __uri__ = uri(State) + "/scalar"

    def __json__(self):
        return to_json(form_of(self))


# User-defined Ops

class Op(Scalar):
    """A callable function."""

    __uri__ = uri(Scalar) + "/op"


class GetOp(Op):
    """A function which can be called via a GET request."""

    __uri__ = uri(Op) + "/get"

    def __call__(self, key=None):
        return OpRef.Get(self, key)


class PutOp(Op):
    """A function which can be called via a PUT request."""

    __uri__ = uri(Op) + "/put"

    def __call__(self, key=None, value=None):
        return OpRef.Put(self, key, value)


class PostOp(Op):
    """A function which can be called via a POST request."""

    __uri__ = uri(Op) + "/post"

    def __call__(self, **params):
        return OpRef.Post(self, **params)


class DeleteOp(Op):
    """A function which can be called via a DELETE request."""

    __uri__ = uri(Op) + "/delete"

    def __call__(self, key=None):
        return OpRef.Delete(self, key)


Op.Get = GetOp
Op.Put = PutOp
Op.Post = PostOp
Op.Delete = DeleteOp


# User-defined object types

class Class(State):
    """A user-defined Tinychain class."""

    __uri__ = uri(State) + "/object/class"

    def __json__(self):
        return {str(uri(Class)): to_json(form_of(self))}


class Instance(State):
    """An instance of a user-defined :class:`Class`."""

    __uri__ = uri(State) + "/object/instance"




"""Generic error types."""

from .util import *


class TinychainError(Exception):
    """An error encountered while executing a transaction."""

    __uri__ = URI("/error")

    def __init__(self, message):
        self.message = message

    def __json__(self):
        return {str(uri(self)): [str(self.message)]}


class BadRequest(TinychainError):
    """Error indicating receipt of a request which is badly-constructed or nonsensical."""

    __uri__ = uri(TinychainError) + "/bad_request"


class Forbidden(TinychainError):
    """Error indicating that the requestor is not authorized to access a requested resource."""

    __uri__ = uri(TinychainError) + "/forbidden"


class MethodNotAllowed(TinychainError):
    """
    Error indicating that the requested resource exists,
    but does not support the request method.
    """

    __uri__ = uri(TinychainError) + "/method_not_allowed"


class NotFound(TinychainError):
    """Error indicating that there is no resource with the requested path and key."""

    __uri__ = uri(TinychainError) + "/not_found"


class NotImplemented(TinychainError):
    """
    Error indicating that the requested functionality does not exist,
    but is planned for implementation in the future.
    """

    __uri__ = uri(TinychainError) + "/not_implemented"


class Unauthorized(TinychainError):
    """Error indicating that the requestor's credentials are missing or invalid."""

    __uri__ = uri(TinychainError) + "/unauthorized"


class UnknownError(TinychainError):
    """An internal error with no handling or recovery logic defined."""

    __uri__ = uri(TinychainError) + "/unknown"




import tinychain.error
import tinychain.host

from tinychain.decorators import *
from tinychain.chain import Chain
from tinychain.cluster import Cluster, write_cluster
from tinychain.ref import *
from tinychain.reflect import Meta
from tinychain.state import Class, Instance, Map, State, Scalar, Tuple
from tinychain.util import *
from tinychain.value import *




"""Reference types."""

from .util import form_of, uri, URI, to_json

class Ref(object):
    """A reference to a :class:`State`."""

    __uri__ = URI("/state/scalar/ref")


class After(Ref):
    """A flow control operator used to delay execution conditionally."""

    __uri__ = uri(Ref) + "/after"

    def __init__(self, when, then):
        self.when = when
        self.then = then

    def __json__(self):
        return {str(uri(self)): to_json([self.when, self.then])}


class Case(Ref):
    """A flow control operator used to branch execution conditionally."""

    __uri__  = uri(Ref) + "/case"

    def __init__(self, cond, switch, case):
        self.cond = cond
        self.switch = switch
        self.case = case

    def __json__(self):
        return {str(uri(self)): to_json([self.cond, self.switch, self.case])}


class If(Ref):
    """A flow control operator used to resolve a :class:`State` conditionally."""

    __uri__ = uri(Ref) + "/if"

    def __init__(self, cond, then, or_else):
        self.cond = cond
        self.then = then
        self.or_else = or_else

    def __json__(self):
        return {str(uri(self)): to_json([self.cond, self.then, self.or_else])}


class OpRef(Ref):
    """A reference to an :class:`Op`."""

    __uri__ = uri(Ref) + "/op"

    def __init__(self, subject, args):
        if isinstance(subject, Ref):
            self.subject = subject
        else:
            self.subject = uri(subject)

        self.args = args

    def __json__(self):
        return {str(self.subject): to_json(self.args)}


class GetOpRef(OpRef):
    """A reference to an instance of :class:`Op.Get`."""

    __uri__ = uri(OpRef) + "/get"

    def __init__(self, subject, key=None):
        if str(uri(subject)).startswith("/state/scalar"):
            OpRef.__init__(self, subject, key)
        else:
            OpRef.__init__(self, subject, (key,))


class PutOpRef(OpRef):
    """A reference to an instance of :class:`Op.Put`."""

    __uri__ = uri(OpRef) + "/put"

    def __init__(self, subject, key, value):
        OpRef.__init__(self, subject, (key, value))


class PostOpRef(OpRef):
    """A reference to an instance of :class:`Op.Post`."""

    __uri__ = uri(OpRef) + "/post"

    def __init__(self, subject, **kwargs):
        OpRef.__init__(self, subject, kwargs)


class DeleteOpRef(OpRef):
    """A reference to an instance of :class:`Op.Delete`."""

    __uri__ = uri(OpRef) + "/delete"

    def __init__(self, subject, key=None):
        OpRef.__init__(self, subject, key)

    def __json__(self):
        return {str(uri(self)): to_json([self.subject, self.args])}


OpRef.Get = GetOpRef
OpRef.Put = PutOpRef
OpRef.Post = PostOpRef
OpRef.Delete = DeleteOpRef




"""Utilities for communicating with a Tinychain host."""

import json
import logging
import os
import pathlib
import requests
import subprocess
import sys
import time
import urllib.parse

from .error import *
from .util import to_json, uri


DEFAULT_PORT = 8702
ENCODING = "utf-8"


class Host(object):
    """A Tinychain host."""

    @staticmethod
    def encode_params(params):
        return urllib.parse.urlencode({k: json.dumps(v) for k, v in params.items()})

    def __init__(self, address):
        self.address = address

    def _handle(self, req):
        response = req()
        status = response.status_code
        response = response.text

        if status == 200:
            return json.loads(response)
        elif status == 204:
            return None
        elif status == 400:
            raise BadRequest(response)
        elif status == 401:
            raise Unauthorized(response)
        elif status == 403:
            raise Forbidden(response)
        elif status == 404:
            raise NotFound(response)
        elif status == 405:
            raise MethodNotAllowed(response)
        elif status == 501:
            raise NotImplemented(response)
        else:
            raise UnknownError(f"HTTP error code {status}: {response}")

    def link(self, path):
        """Return a link to the given path at this host."""

        return "http://{}{}".format(self.address, path)

    def get(self, path, key=None, auth=None):
        """Execute a GET request."""

        url = self.link(path)
        headers = auth_header(auth)
        if key:
            key = json.dumps(to_json(key)).encode(ENCODING)
            request = lambda: requests.get(url, params={"key": key}, headers=headers)
        else:
            request = lambda: requests.get(url, headers=headers)

        return self._handle(request)

    def put(self, path, key, value, auth=None):
        """Execute a PUT request."""

        url = self.link(path)
        headers = auth_header(auth)
        key = json.dumps(to_json(key)).encode(ENCODING)
        value = json.dumps(to_json(value)).encode(ENCODING)
        request = lambda: requests.put(url, params={"key": key}, data=value, headers=headers)

        return self._handle(request)

    def post(self, path, data, auth=None):
        """Execute a POST request."""

        url = self.link(path)
        data = json.dumps(to_json(data)).encode(ENCODING)
        headers = auth_header(auth)
        request = lambda: requests.post(url, data=data, headers=headers)

        return self._handle(request)

    def delete(self, path, key=None, auth=None):
        """Execute a DELETE request."""

        url = self.link(path)
        headers = auth_header(auth)
        if key:
            key = json.dumps(to_json(key)).encode(ENCODING)
            request = lambda: requests.delete(url, params={"key": key}, headers=headers)
        else:
            request = lambda: requests.delete(url, headers=headers)

        return self._handle(request)

    def resolve(self, state, auth=None):
        """Resove the given state."""

        return self.post("/transact/execute", state, auth)


class Local(Host):
    """A local Tinychain host."""

    ADDRESS = "127.0.0.1"
    STARTUP_TIME = 0.5

    def __init__(self,
            path,
            workspace,
            data_dir=None,
            clusters=[],
            port=DEFAULT_PORT,
            log_level="warn",
            force_create=False):

        # set _process first so it's available to __del__ in case of an exception
        self._process = None

        if not int(port) or int(port) < 0:
            raise ValueError(f"invalid port: {port}")

        if clusters and data_dir is None:
            raise ValueError("Hosting a cluster requires specifying a data_dir")

        maybe_create_dir(workspace, force_create)

        if data_dir:
            maybe_create_dir(data_dir, force_create)

        address = "{}:{}".format(self.ADDRESS, port)
        Host.__init__(self, address)

        args = [
            path,
            f"--http_port={port}",
            f"--log_level={log_level}",
        ]

        if data_dir:
            args.append(f"--data_dir={data_dir}")

        args.extend([f"--cluster={cluster}" for cluster in clusters])

        self._process = subprocess.Popen(args)
        time.sleep(self.STARTUP_TIME)

        if self._process is None or self._process.poll() is not None:
            raise RuntimeError(f"Tinychain process at {self.address} crashed on startup")
        else:
            logging.info(f"new instance running at {self.address}: {workspace}")

    def stop(self):
        """Shut down this host."""

        logging.info(f"Shutting down Tinychain host {self.address}")
        self._process.terminate()
        self._process.wait()
        logging.info(f"Host {self.address} shut down successfully")

    def __del__(self):
        if self._process:
            self.stop()


def auth_header(token):
    return {"Authorization": f"Bearer {token}"} if token else {}


def maybe_create_dir(path, force):
    path = pathlib.Path(path)
    if path.exists() and path.is_dir():
        return
    elif force:
        os.makedirs(path)
    else:
        raise RuntimeError(f"no directory at {path}")




"""Base class of a hosted service."""

import inspect
import logging

from .decorators import *
from .ref import OpRef
from .reflect import Meta
from .state import Op, State, Tuple
from .util import form_of, ref as get_ref, uri, URI, to_json

class Cluster(object, metaclass=Meta):
    """A hosted Tinychain service."""

    @classmethod
    def __use__(cls):
        instance = cls()
        for name, attr in inspect.getmembers(instance):
            if name.startswith('_'):
                continue

            if isinstance(attr, MethodStub):
                setattr(instance, name, attr.method(instance, name))
            elif inspect.isclass(attr) and issubclass(attr, State):
                if not str(uri(cls)).startswith(str(uri(attr))):
                    logging.warning(f"cluster at {uri(cls)} serves class at {uri(attr)}")

                @get_method
                def ctr(self, txn, form) -> attr:
                    return attr(form)

                setattr(instance, name, ctr.method(instance, name))

        return instance

    def __init__(self, form=None):
        self.__form__ = form if form else uri(self)
        self._configure()

    def _configure(self):
        """Initialize this `Cluster`'s :class:~`chain.Chain`s."""
        pass

    def authorize(self, scope):
        """Raise an error if the current transaction has not been granted the given scope."""

        return OpRef.Get(uri(self) + "/authorize", scope)

    def grant(self, scope, op: Op, context={}):
        """Execute the given `op` after granting it the given `scope`."""

        params = {
            "scope": scope,
            "op": op,
        }

        if context:
            params["context"] = context

        return OpRef.Post(uri(self) + "/grant", **params)

    @put_method
    def install(self, txn, cluster_link: URI, scopes: Tuple):
        """Trust the cluster at the given link to grant the given scopes."""
        pass


def write_cluster(cluster, config_path, overwrite=False):
    """Write the configuration of the given :class:`Cluster` to the given path."""

    import json
    import pathlib

    config = {str(uri(cluster)): to_json(form_of(cluster))}
    config_path = pathlib.Path(config_path)
    if config_path.exists() and not overwrite:
        with open(config_path) as f:
            try:
                if json.load(f) == config:
                    return
            except json.decoder.JSONDecodeError as e:
                logging.warning(f"invalid JSON at {config_path}: {e}")
                pass

        raise RuntimeError(f"There is already an entry at {config_path}")
    else:
        import os

        if not config_path.parent.exists():
            os.makedirs(config_path.parent)

        with open(config_path, 'w') as cluster_file:
            cluster_file.write(json.dumps(config, indent=4))




""":class:`Value` types such as :class:`Nil`, :class:`Number`, and :class:`String`."""

from .ref import OpRef
from .reflect import Meta
from .state import Scalar
from .util import uri


# Scalar value types

class Value(Scalar, metaclass=Meta):
    """A scalar `Value` which supports equality and collation."""

    __uri__ = uri(Scalar) + "/value"

    def __eq__(self, other):
        return self.eq(other)

    def __ne__(self, other):
        return self.ne(other)

    def eq(self, other):
        """Returns `true` if `self` is equal to `other`."""

        return _get_op("eq", self, other)

    def ne(self, other):
        """Returns `true` if `self` is not equal to `other`."""

        return _get_op("ne", self, other)


class Nil(Value):
    """A Tinychain `None` Value."""

    __uri__ = uri(Value) + "/none"


class String(Value):
    """A string."""

    __uri__ = uri(Value) + "/string"


# Numeric types

class Number(Value):
    """A numeric :class:`Value`."""

    __uri__ = uri(Value) + "/number"

    def __add__(self, other):
        return self.add(other)

    def __div__(self, other):
        return self.div(other)

    def __gt__(self, other):
        return self.gt(other)

    def __ge__(self, other):
        return self.gte(other)

    def __lt__(self, other):
        return self.lt(other)

    def __le__(self, other):
        return self.lte(other)

    def __mul__(self, other):
        return self.mul(other)

    def __radd__(self, other):
        return self.add(other)

    def __rmul__(self, other):
        return self.mul(other)

    def __sub__(self, other):
        return self.sub(other)

    def __truediv__(self, other):
        return self.div(other)

    def add(self, other):
        """Return the sum of `self` and `other`."""

        return _get_op("add", self, other)

    def div(self, other):
        """Return the quotient of `self` and `other`."""

        return _get_op("div", self, other)

    def gt(self, other):
        """Return true if `self` is greater than `other`."""

        return _get_op("gt", self, other, Bool)

    def gte(self, other):
        """Return true if `self` is greater than or equal to `other`."""

        return _get_op("gte", self, other, Bool)

    def lt(self, other):
        """Return true if `self` is less than `other`."""

        return _get_op("lt", self, other, Bool)

    def lte(self, other):
        """Return true if `self` is less than or equal to `other`."""

        return _get_op("lte", self, other, Bool)

    def mul(self, other):
        """Return the product of `self` and `other`."""

        return _get_op("mul", self, other)

    def sub(self, other):
        """Return the difference between `self` and `other`."""

        return _get_op("sub", self, other)


def _get_op(name, subject, key, dtype=Number):
    return dtype(OpRef.Get(uri(subject).append(name), key))


class Bool(Number):
    """A boolean :class:`Value`."""

    __uri__ = uri(Number) + "/bool"




"""Utility data structures and functions."""

import inspect

from collections import OrderedDict


class Context(object):
    """A transaction context."""

    def __init__(self, context=None):
        object.__setattr__(self, "form", OrderedDict())

        if context:
            if isinstance(context, self.__class__):
                for name, value in context.form.items():
                    setattr(self, name, value)
            else:
                for name, value in dict(context).items():
                    setattr(self, name, value)

    def __add__(self, other):
        concat = Context(self)

        if isinstance(other, self.__class__):
            for name, value in other.form.items():
                setattr(concat, name, value)
        else:
            for name, value in dict(other).items():
                setattr(concat, name, value)

        return concat

    def __getattr__(self, name):
        if name in self.form:
            value = self.form[name]
            if isinstance(value, type):
                from .state import Class
                return Class(URI(name))
            elif hasattr(value, "__ref__"):
                return ref(value, name)
            else:
                return URI(name)
        else:
            raise ValueError(f"Context has no such value: {name}")

    def __json__(self):
        return to_json(list(self.form.items()))

    def __setattr__(self, name, value):
        if name in object.__getattribute__(self, "form"):
            raise ValueError(f"Context already has a value named {name}")
        else:
            self.form[name] = value


def form_of(state):
    """Return the form of the given state."""

    if hasattr(state, "__form__"):
        if callable(state.__form__) and not inspect.isclass(state.__form__):
            return state.__form__()
        else:
            return state.__form__
    else:
        raise ValueError(f"{state} has no form")


def ref(subject, name):
    """Return a named reference to the given state."""

    if hasattr(subject, "__ref__"):
        return subject.__ref__(name)
    else:
        return subject


class URI(object):
    """
    An absolute or relative link to a Tinychain state.
    
    Examples:
        `URI("http://example.com/myservice/value_name")`
        `URI("$other_state/method_name")`
        `URI("/state/scalar/value/none")`
    """

    def __init__(self, root, path=[]):
        assert root is not None

        self._root = root
        self._path = path

    def append(self, name):
        """
        Append a segment to this `URI`.
        
        Example:
            `value = OpRef.Get(URI("http://example.com/myapp").append("value_name"))`
        """ 
        return URI(str(self), [name])

    def host(self):
        """Return the host segment of this `URI`, if present."""

        if "://" not in self._root:
            return None

        start = self._root.index("://") + 3
        end = (
            self._root.index(':', start) if ':' in self._root[start:]
            else self._root.index('/', start))

        if end > start:
            return self._root[start:end]
        else:
            return self._root[start:]

    def path(self):
        """Return the path segment of this `URI`."""

        if "://" not in self._root:
            return self._root + "/".join(self._path)

        start = self._root.index("://")
        start = self._root.index('/', start + 3)

        if self._path:
            return self._root[start:] + "/".join(self._path)
        else:
            return self._root[start:]

    def port(self):
        """Return the port of this `URI`, if present."""

        prefix = self.protocol() + "://" if self.protocol() else ""
        prefix += self.host() if self.host() else ""
        if prefix and self._root[len(prefix)] == ':':
            end = self._root.index('/', len(prefix))
            return int(self._root[len(prefix) + 1:end])

    def protocol(self):
        """Return the protocol of this `URI` (e.g. "http"), if present."""

        if "://" in self._root:
            i = self._root.index("://")
            if i > 0:
                return self._root[:i]

    def __add__(self, other):
        return URI(str(self) + other)

    def __radd__(self, other):
        return URI(other) + str(self)

    def __json__(self):
        return {str(self): []}

    def __str__(self):
        root = str(self._root)
        if root.startswith('/') or root.startswith('$') or "://" in root:
            pass
        else:
            root = f"${root}"

        if self._path:
            path = "/".join(self._path)
            return f"{root}/{path}"
        else:
            return root


def uri(subject):
    """Return the `URI` of the given state."""

    if hasattr(subject, "__uri__"):
        return subject.__uri__
    elif isinstance(subject, URI):
        return subject
    elif hasattr(type(subject), "__uri__"):
        return uri(type(subject))
    else:
        raise AttributeError(f"{subject} has no URI")


def use(cls):
    """Return an instance of the given class with callable methods."""

    if hasattr(cls, "__use__"):
        return cls.__use__()
    else:
        cls()


def to_json(obj):
    """Return a JSON-encodable representation of the given state or reference."""

    if inspect.ismethod(obj):
        if not hasattr(obj, "__json__"):
            raise ValueError(
                f"Python method {obj} is not JSON serializable; "
                + "try using a decorator like @get_method")

    if inspect.isclass(obj):
        if hasattr(type(obj), "__json__"):
            return type(obj).__json__(obj)

    if hasattr(obj, "__json__"):
        return obj.__json__()
    elif isinstance(obj, list) or isinstance(obj, tuple):
        return [to_json(i) for i in obj]
    elif isinstance(obj, dict):
        return {to_json(k): to_json(v) for k, v in obj.items()}
    else:
        return obj




import inspect

from pydoc import locate

from . import error
from .ref import OpRef
from .state import Class, Op as OpDef, Scalar, State
from .util import *


def gen_headers(instance):
    for name, attr in inspect.getmembers(instance):
        if name.startswith('_'):
            continue

        if isinstance(attr, MethodStub):
            setattr(instance, name, attr.method(instance, name))


class Meta(type):
    """The metaclass of a :class:`State`."""

    def __form__(cls):
        mro = cls.mro()
        if len(mro) < 2:
            raise ValueError("Tinychain class must extend a subclass of State")

        parent_members = dict(inspect.getmembers(mro[1](URI("self"))))

        class Header(cls):
            pass

        header = Header(URI("self"))
        instance = cls(URI("self"))

        for name, attr in inspect.getmembers(instance):
            if name.startswith('_'):
                continue

            if isinstance(attr, MethodStub):
                setattr(header, name, attr.method(instance, name))
            elif isinstance(attr, State):
                setattr(header, name, type(attr)(URI(f"self/{name}")))
            else:
                setattr(header, name, attr)

        form = {}
        for name, attr in inspect.getmembers(instance):
            if name.startswith('_'):
                continue
            elif name in parent_members:
                if attr is parent_members[name] or attr == parent_members[name]:
                    continue
                elif hasattr(attr, "__code__") and hasattr(parent_members[name], "__code__"):
                    if attr.__code__ is parent_members[name].__code__:
                        continue

            if isinstance(attr, MethodStub):
                form[name] = to_json(attr.method(header, name))
            else:
                form[name] = attr

        return form

    def __json__(cls):
        mro = cls.mro()
        if len(mro) < 2:
            raise ValueError("Tinychain class must extend a subclass of State")

        parent = mro[1]
        return {str(uri(parent)): to_json(form_of(cls))}


class MethodStub(object):
    def __init__(self, dtype, form):
        self.dtype = dtype
        self.form = form

    def __call__(self, *args, **kwargs):
        raise RuntimeError(
            "cannot call a MethodStub; use tc.use(<class>) for callable method references")

    def method(self, header, name):
        return self.dtype(header, self.form, name)


class Method(object):
    __uri__ = uri(OpDef)

    def __init__(self, header, form, name):
        self.header = header
        self.form = form
        self.name = name

    def __json__(self):
        return {str(uri(self)): to_json(form_of(self))}

    def dtype(self):
        return self.__class__.__name__


class GetMethod(Method):
    __uri__ = uri(OpDef.Get)

    def __call__(self, key=None):
        from .value import Nil
        rtype = inspect.signature(self.form).return_annotation
        rtype = resolve_class(self.form, rtype, Nil)
        return rtype(OpRef.Get(uri(self.header).append(self.name), key))

    def __form__(self):
        sig = inspect.signature(self.form)
        parameters = list(sig.parameters.items())

        if len(parameters) < 1 or len(parameters) > 3:
            raise ValueError(f"{self.dtype()} takes 1-3 arguments: (self, cxt, key)")

        args = [self.header]

        cxt = Context()
        if len(parameters) > 1:
            args.append(cxt)

        key_name = "key"
        if len(parameters) == 3:
            from .value import Value
            key_name, param = parameters[2]
            dtype = resolve_class(self.form, param.annotation, Value)
            args.append(dtype(URI(key_name)))

        cxt._return = self.form(*args) # populate the Context
        return (key_name, cxt)


class PutMethod(Method):
    __uri__ = uri(OpDef.Put)

    def __call__(self, key, value):
        return OpRef.Put(uri(self.header) + "/" + self.name, key, value)

    def __form__(self):
        sig = inspect.signature(self.form)
        parameters = list(sig.parameters.items())

        if len(parameters) not in [1, 2, 4]:
            raise ValueError("{self.dtype()} has one, two, or four arguments: "
                + "(self, cxt, key, value)")

        args = [self.header]

        cxt = Context()
        if len(parameters) > 1:
            args.append(cxt)

        key_name = "key"
        value_name = "value"
        if len(parameters) == 4:
            from .value import Value
            key_name, param = parameters[2]
            dtype = resolve_class(self.form, param.annotation, Value)
            args.append(dtype(URI(key_name)))

            value_name, param = parameters[3]
            dtype = resolve_class(self.form, param.annotation)
            args.append(dtype(URI(value_name)))

        cxt._return = self.form(*args)
        return (key_name, value_name, cxt)


class PostMethod(Method):
    __uri__ = uri(OpDef.Post)

    def __call__(self, **params):
        from .value import Nil

        rtype = inspect.signature(self.form).return_annotation
        rtype = resolve_class(self.form, rtype, Nil)
        return rtype(OpRef.Post(uri(self.header).append(self.name), **params))

    def __form__(self):
        sig = inspect.signature(self.form)
        parameters = list(sig.parameters.items())

        if len(parameters) == 0:
            raise ValueError("{self.dtype()} has at least one argment: "
                + "(self, cxt, name1=val1, ...)")

        args = [self.header]

        cxt = Context()
        if len(parameters) > 1:
            args.append(cxt)

        kwargs = {}
        for name, param in parameters[2:]:
            dtype = resolve_class(self.form, param.annotation)
            kwargs[name] = dtype(URI(name))

        cxt._return = self.form(*args, **kwargs)
        return cxt


class DeleteMethod(Method):
    __uri__ = uri(OpDef.Delete)

    def __call__(self, key=None):
        return OpRef.Delete(uri(self.header).append(self.name), key)

    def __form__(self):
        return GetMethod.__form__(self)


Method.Get = GetMethod
Method.Put = PutMethod
Method.Post = PostMethod
Method.Delete = DeleteMethod


class Op(object):
    __uri__ = uri(OpDef)

    def __init__(self, form):
        self.form = form

    def __json__(self):
        return {str(uri(self)): to_json(form_of(self))}

    def dtype(self):
        return self.__class__.__name__


class GetOp(Op):
    __uri__ = uri(OpDef.Get)

    def __call__(self, key=None):
        return OpRef.Get(uri(self), key)

    def __form__(self):
        sig = inspect.signature(self.form)
        parameters = list(sig.parameters.items())

        if len(parameters) > 2:
            raise ValueError(f"{self.dtype()} takes 0-2 arguments: (cxt, key)")

        args = []

        cxt = Context()
        if len(parameters):
            args.append(cxt)

        key_name = "key"
        if len(parameters) == 2:
            from .value import Value
            key_name, param = parameters[1]
            dtype = resolve_class(self.form, param.annotation, Value)
            args.append(dtype(URI(key_name)))

        cxt._return = self.form(*args) # populate the Context
        return (key_name, cxt)

    def __ref__(self, name):
        return OpDef.Get(URI(name))


class PutOp(Op):
    __uri__ = uri(OpDef.Put)

    def __form__(self):
        sig = inspect.signature(self.form)
        parameters = list(sig.parameters.items())

        if len(parameters) not in [0, 1, 3]:
            raise ValueError("{self.dtype()} has 0, 1, or 3 arguments: (cxt, key, value)")

        args = [self.header]

        cxt = Context()
        if len(parameters):
            args.append(cxt)

        key_name = "key"
        value_name = "value"
        if len(parameters) == 3:
            from .value import Value
            key_name, param = parameters[1]
            dtype = resolve_class(self.form, param.annotation, Value)
            args.append(dtype(URI(key_name)))

            value_name, param = parameters[2]
            dtype = resolve_class(self.form, param.annotation)
            args.append(dtype(URI(value_name)))

        cxt._return = self.form(*args)
        return (key_name, value_name, cxt)

    def __ref__(self, name):
        return OpDef.Put(URI(name))


class PostOp(Op):
    __uri__ = uri(OpDef.Post)

    def __form__(self):
        sig = inspect.signature(self.form)
        parameters = list(sig.parameters.items())

        args = []

        cxt = Context()
        if len(parameters):
            args.append(cxt)

        kwargs = {}
        for name, param in parameters[1:]:
            dtype = resolve_class(self.form, param.annotation)
            kwargs[name] = dtype(URI(name))

        cxt._return = self.form(*args, **kwargs)
        return cxt

    def __ref__(self, name):
        return OpDef.Post(URI(name))


class DeleteOp(Op):
    __uri__ = uri(OpDef.Delete)

    def __form__(self):
        return GetOp.__form__(self)


    def __ref__(self, name):
        return OpDef.Delete(URI(name))


Op.Get = GetOp
Op.Put = PutOp
Op.Post = PostOp
Op.Delete = DeleteOp


def resolve_class(subject, annotation, default=State):
    if annotation == inspect.Parameter.empty:
        return default
    elif inspect.isclass(annotation):
        return annotation

    classpath = f"{subject.__module__}.{annotation}"
    resolved = locate(classpath)
    if resolved is None:
        raise ValueError(f"unable to resolve class {classpath}")
    else:
        return resolved




"""Data structures responsible for keeping track of mutations to a :class:`Value`"""

from .ref import OpRef
from .reflect import Meta
from .state import Scalar, State
from .util import *


class Chain(State, metaclass=Meta):
    """
    Data structure responsible for keeping track of mutations to a :class:`Value`.
    """

    __uri__ = uri(State) + "/chain"

    def set(self, value):
        """Update the value of this `Chain`."""

        return OpRef.Put(uri(self).append("subject"), None, value)

    def subject(self, key=None):
        """Return a reference to the subject of this `Chain`."""

        return OpRef.Get(uri(self).append("subject"), key)


class SyncChain(Chain):
    """
    A :class:`Chain` which keeps track of only the current transaction's operations,
    in order to recover from a transaction failure (e.g. if the host crashes).
    """

    __uri__ = uri(Chain) + "/sync"


Chain.Sync = SyncChain

